import pytest

import hoa_tools.dataset
import hoa_tools.registration
import hoa_tools.voi
from hoa_tools.types import ArrayCoordinate


def test_transform_voi() -> None:
    overview = hoa_tools.dataset.get_dataset(
        "S-20-29_brain_complete-organ_25.33um_bm05"
    )
    child = hoa_tools.dataset.get_dataset("S-20-29_brain_VOI-04_6.5um_bm05")

    child_voi = hoa_tools.voi.VOI(
        dataset=child,
        downsample_level=0,
        lower_corner=ArrayCoordinate(x=3434, y=2060, z=2656),
        size=ArrayCoordinate(x=256, y=256, z=128),
    )

    overview_voi = child_voi.transform_to(overview)
    assert overview_voi.dataset.name == "S-20-29_brain_complete-organ_25.33um_bm05"
    assert overview_voi.downsample_level == 0
    assert overview_voi.lower_corner == ArrayCoordinate(x=2975, y=1689, z=4316)
    assert overview_voi.size == ArrayCoordinate(x=69, y=69, z=33)


def test_inverse_registration() -> None:
    overview = hoa_tools.dataset.get_dataset(
        "S-20-29_brain_complete-organ_25.33um_bm05"
    )
    zoom = hoa_tools.dataset.get_dataset("S-20-29_brain_VOI-04_6.5um_bm05")

    transform = hoa_tools.registration.Inventory.get_registration(
        source_dataset=zoom, target_dataset=overview
    )
    transform_inv = hoa_tools.registration.Inventory.get_registration(
        source_dataset=overview, target_dataset=zoom
    )

    assert transform.GetParameters() == (
        0.020069870800809003,
        0.0,
        0.0,
        92062.68586067778,
        30372.305582420606,
        52529.27516774854,
        1.0000012996591003,
    )

    # Check transform roundtrip
    point = (0, 0, 0)
    assert transform_inv.TransformPoint(transform.TransformPoint(point)) == point


def test_transform_path() -> None:
    # Test getting transform between two datasets that aren't directly registered
    zoom1 = hoa_tools.dataset.get_dataset("S-20-29_brain_VOI-04_6.5um_bm05")
    zoom2 = hoa_tools.dataset.get_dataset("S-20-29_brain_VOI-05_6.5um_bm05")

    hoa_tools.registration.Inventory.get_registration(
        source_dataset=zoom1, target_dataset=zoom2
    )


def test_no_transform_path() -> None:
    # Test getting transform between two datasets that aren't directly registered
    d1 = hoa_tools.dataset.get_dataset("S-20-29_brain_VOI-04_6.5um_bm05")
    d2 = hoa_tools.dataset.get_dataset(
        "LADAF-2020-27_spleen_complete-organ_25.08um_bm05"
    )

    assert (d1, d2) not in hoa_tools.registration.Inventory

    with pytest.raises(
        ValueError,
        match=(
            "No registration path between S-20-29_brain_VOI-04_6.5um_bm05 and "
            "LADAF-2020-27_spleen_complete-organ_25.08um_bm05"
        ),
    ):
        hoa_tools.registration.Inventory.get_registration(
            source_dataset=d1, target_dataset=d2
        )


def test_registered() -> None:
    # Test getting transform between two datasets that aren't directly registered
    d = hoa_tools.dataset.get_dataset("S-20-29_brain_VOI-04_6.5um_bm05")
    registered = d.get_registered()
    assert {d.name for d in registered} == {
        "S-20-29_brain_complete-organ_25.33um_bm05",
        "S-20-29_brain_VOI-04_2.5um_bm05",
        "S-20-29_brain_VOI-03_6.5um_bm05",
        "S-20-29_brain_VOI-01_2.5um_bm05",
        "S-20-29_brain_VOI-03_2.5um_bm05",
        "S-20-29_brain_VOI-01b_6.5um_bm05",
        "S-20-29_brain_VOI-02_2.5um_bm05",
        "S-20-29_brain_VOI-01_6.5um_bm05",
        "S-20-29_brain_VOI-02_6.5um_bm05",
        "S-20-29_brain_VOI-05_2.5um_bm05",
        "S-20-29_brain_VOI-04_6.5um_bm05",
        "S-20-29_brain_VOI-05_6.5um_bm05",
    }


def test_transform_zoom_to_zoom() -> None:
    zoom1 = hoa_tools.dataset.get_dataset("LADAF-2021-64_heart_VOI-07_6.51um_bm18")
    zoom2 = hoa_tools.dataset.get_dataset("LADAF-2021-64_heart_VOI-7.1_2.26um_bm18")

    voi = hoa_tools.voi.VOI(
        dataset=zoom1,
        downsample_level=0,
        lower_corner={"x": 3434, "y": 2060, "z": 2656},
        size={"x": 256, "y": 256, "z": 128},
    )

    voi.transform_to(zoom2)



"""
Tools for working with individual datasets.

The core class representing a single dataset is [`Dataset`][hoa_tools.dataset.Dataset].
This inherits from [`HOAMetadata`][hoa_tools.metadata.HOAMetadata],
which stores all the metadata for a given dataset.

[`Dataset`][hoa_tools.dataset.Dataset] objects are not designed to be created by users.
To get a [`Dataset`][hoa_tools.dataset.Dataset], use the
[`get_dataset`][hoa_tools.dataset.get_dataset] function in this module.
"""

import warnings
from functools import cached_property
from pathlib import Path
from typing import Literal

import dask.array.core
import gcsfs
import networkx as nx
import numpy as np
import xarray as xr
import zarr.core
import zarr.n5
import zarr.storage

from hoa_tools._n5 import N5FSStore
from hoa_tools.metadata import HOAMetadata
from hoa_tools.types import PhysicalCoordinate

__all__ = ["Dataset", "get_dataset"]


_DATASETS: dict[str, "Dataset"]


class Dataset(HOAMetadata):
    """
    An individual Human Organ Atlas dataset.
    """

    def __str__(self) -> str:
        return f"Dataset(name={self.name})"

    def __hash__(self) -> int:
        return hash(self.name)

    def _organ_str(self) -> str:
        """
        Get name of organ, with organ context appended if present.
        """
        organ_str = str(self.sample.organ)
        if self.sample.organ_context:
            organ_str += "_" + self.sample.organ_context
        return organ_str

    def _voxel_size_str(self) -> str:
        # Make sure that voxel size includes a .0 if an integer value.
        return str(float(self.data.voxel_size_um))

    @property
    def is_full_organ(self) -> bool:
        """
        Whether this dataset contains the whole organ or not.
        """
        return self.voi.startswith("complete")

    @property
    def is_zoom(self) -> bool:
        """
        Whether this dataset is a zoom or not.
        """
        return not self.is_full_organ

    def get_children(self) -> list["Dataset"]:
        """
        Get child dataset(s).

        Child datasets are high-resolution zooms of a full-organ dataset.

        Notes
        -----
        For full-organ datasets, this returns an empty list.

        """
        children = [
            d
            for d in _DATASETS.values()
            if (
                d.donor.id == self.donor.id
                and d.sample.organ == self.sample.organ
                and d.scan.beamline == self.scan.beamline
                and d.is_zoom
            )
        ]
        return sorted(children, key=lambda c: c.name)

    def get_parents(self) -> list["Dataset"]:
        """
        Get parent dataset(s).

        Parent datasets are full-organ datasets from which a zoom dataset is taken from.

        Notes
        -----
        For zoom datasets, this returns an empty list.

        """
        parents = [
            d
            for d in _DATASETS.values()
            if (
                d.donor.id == self.donor.id
                and d.sample.organ == self.sample.organ
                and d.scan.beamline == self.scan.beamline
                and d.is_full_organ
            )
        ]
        return sorted(parents, key=lambda c: c.name)

    def get_registered(self) -> set["Dataset"]:
        """
        Get a set of all datasets that are registered (even indirectly) to this dataset.
        """
        import hoa_tools.registration  # noqa: PLC0415

        dataset_names = nx.node_connected_component(
            hoa_tools.registration.Inventory._graph.to_undirected(),  # noqa: SLF001
            self.name,
        )
        return {get_dataset(name) for name in dataset_names}

    @property
    def _remote_fmt(self) -> Literal["n5", "zarr"]:
        if self.data.gcs_url.startswith("n5://"):
            return "n5"
        if self.data.gcs_url.startswith("zarr://"):
            return "zarr"
        raise RuntimeError("URL must start with n5:// or zarr://")

    @cached_property
    def _remote_store(self) -> zarr.Group:
        """
        Remote data store.
        """
        gcs_url = self.data.gcs_url
        gcs_path = gcs_url.removeprefix("n5://gs://").removeprefix("zarr://gs://")

        # n5://gs://ucl-hip-ct-35a68e99feaae8932b1d44da0358940b/S-20-29/heart/2.5um_VOI-01_bm05/
        bucket, path = gcs_path.split("/", maxsplit=1)
        fs = gcsfs.GCSFileSystem(project=bucket, token="anon", access="read_only")  # noqa: S106
        if self._remote_fmt == "n5":
            store = N5FSStore(url=bucket, fs=fs, mode="r")
        elif self._remote_fmt == "zarr":
            store = zarr.storage.FSStore(url=bucket, fs=fs, mode="r")

        return zarr.open_group(store, mode="r", path=path)

    def _remote_array(self, *, downsample_level: int) -> zarr.core.Array:
        """
        Get an object representing the data array in the remote Google Cloud Store.
        """
        if not downsample_level >= 0:
            msg = "level must be >= 0"
            raise ValueError(msg)

        if self._remote_fmt == "n5":
            key = f"s{downsample_level}"
        else:
            key = f"{downsample_level}"
        return self._remote_store[key]

    def data_array(self, *, downsample_level: int) -> xr.DataArray:
        """
        Get a DataArray representing the array for this image.
        """
        remote_array = self._remote_array(downsample_level=downsample_level)
        dask_array = dask.array.core.from_array(  # type: ignore[no-untyped-call]
            remote_array, chunks=remote_array.chunks
        )
        if self._remote_fmt == "zarr":
            dask_array = dask_array.T

        spacing = self.data.voxel_size_um * 2**downsample_level
        return xr.DataArray(
            dask_array,
            name=self.name,
            dims=["z", "y", "x"],
            coords={
                "z": xr.DataArray(
                    data=(np.arange(dask_array.shape[0]) * spacing),
                    dims=["z"],
                    attrs={"units": "μm"},
                ),
                "y": xr.DataArray(
                    data=(np.arange(dask_array.shape[1]) * spacing),
                    dims=["y"],
                    attrs={"units": "μm"},
                ),
                "x": xr.DataArray(
                    data=(np.arange(dask_array.shape[2]) * spacing),
                    dims=["x"],
                    attrs={"units": "μm"},
                ),
            },
        )


def _load_datasets_from_files(data_dir: Path) -> dict[str, Dataset]:
    """
    Load dataset metadtata files.
    """
    datasets = {
        f.stem: Dataset.model_validate_json(f.read_text())
        for f in (data_dir).glob("*.json")
    }
    if len(datasets) == 0:
        raise FileNotFoundError(
            f"Did not find any dataset metadata files at {data_dir}"  # noqa: EM102
        )
    return datasets


def get_dataset(name: str) -> Dataset:
    """
    Get a dataset from its name.

    The name of datasets can be looked up using the `hoa_tools.inventory` module.
    """
    return _DATASETS[name]


def change_metadata_directory(data_dir: Path) -> None:
    """
    Update available datasets from another directory of metadata files.

    Designed for internal project members to load metadata files that aren't yet public.
    """
    global _DATASETS  # noqa: PLW0603
    _DATASETS = _load_datasets_from_files(data_dir)
    _populate_registrations_from_metadata(_DATASETS)


def _populate_registrations_from_metadata(datasets: dict[str, Dataset]) -> None:
    from hoa_tools.registration import Inventory, build_transform  # noqa: PLC0415

    Inventory._clear()  # noqa: SLF001
    for dataset_name in datasets:
        dataset = _DATASETS[dataset_name]
        if (registration := dataset.registration) is not None:
            source_dataset = _DATASETS[registration.source_dataset]
            if registration.target_dataset not in _DATASETS:
                # Some datasets don't have their parent datasets release yet - only warn
                # if we're expecting a parent dataset
                if registration.source_dataset not in [
                    "A129_lung_VOI-02_2.0um_bm18"
                ] and not registration.source_dataset.startswith("LADAF-2021-17_brain"):
                    warnings.warn(
                        f"Did not find target dataset {registration.target_dataset} "
                        f"in dataset inventory. "
                        f"Not adding {registration.source_dataset} "
                        "to registration inventory.",
                        stacklevel=1,
                    )
                continue
            target_dataset = _DATASETS[registration.target_dataset]
            transform = build_transform(
                translation=PhysicalCoordinate(
                    x=registration.translation[0] * target_dataset.data.voxel_size_um,
                    y=registration.translation[1] * target_dataset.data.voxel_size_um,
                    z=registration.translation[2] * target_dataset.data.voxel_size_um,
                ),
                rotation_deg=registration.rotation,
                scale=registration.scale
                * target_dataset.data.voxel_size_um
                / source_dataset.data.voxel_size_um,
            )
            Inventory.add_registration(
                source_dataset=_DATASETS[registration.source_dataset],
                target_dataset=_DATASETS[registration.target_dataset],
                transform=transform,
            )


_META_DIR = Path(__file__).parent / "data" / "metadata" / "metadata"
try:
    change_metadata_directory(_META_DIR)
except FileNotFoundError as e:
    raise ImportError(
        "Did not find any dataset metadata files. "
        "This means there is something wrong with the hoa-tools installation.\n"
        "Please report this at https://github.com/HumanOrganAtlas/hoa-tools\n"
        "If you are installing from source, you might need to "
        "initialise and fetch git submodules. "
    ) from e



# generated by datamodel-codegen:
#   filename:  metadata-schema.json
#   timestamp: 2025-10-22T15:38:14+00:00

from __future__ import annotations

from datetime import date as date_aliased
from typing import Annotated, Literal

from pydantic import BaseModel, ConfigDict, Field, RootModel


class Orcid(RootModel[str]):
    root: Annotated[
        str,
        Field(
            description="ORCID iD",
            pattern="^\\d{4}-\\d{4}-\\d{4}-\\d{3}[0-9X]{1}$",
            title="Orcid",
        ),
    ]


class Contributor(BaseModel):
    first_name: Annotated[str, Field(title="First Name")]
    last_name: Annotated[str, Field(title="Last Name")]
    orcid: Annotated[Orcid | None, Field(description="ORCID iD", title="Orcid")] = None
    roles: Annotated[
        list[
            Literal[
                "Conceptualization",
                "Data curation",
                "Formal analysis",
                "Funding acquisition",
                "Investigation",
                "Methodology",
                "Project administration",
                "Resources",
                "Software",
                "Supervision",
                "Validation",
                "Vlisualization",
                "Writing - original draft",
                "Writing - review & editing",
            ]
        ],
        Field(
            description="Contributor role(s), following the Contributor Role Taxonomy (CReditT)",
            title="Roles",
        ),
    ]


class Data(BaseModel):
    shape: Annotated[
        list,
        Field(
            description="Array shape for dataset.",
            max_length=3,
            min_length=3,
            title="Shape",
        ),
    ]
    voxel_size_um: Annotated[
        float,
        Field(
            description="Isotropic size of a single voxel, in micrometers.",
            gt=0.0,
            title="Voxel size",
        ),
    ]
    gcs_url: Annotated[
        str,
        Field(
            description="URL to data in Google Cloud Storage. Starts with either n5:// or zarr:// to specify whether data is in N5 or OME-Zarr 0.4 format.",
            title="Google cloud storage URL",
        ),
    ]


class Age(RootModel[int]):
    root: Annotated[int, Field(description="Age in years at death.", ge=0, title="Age")]


class Weight(RootModel[float]):
    root: Annotated[
        float, Field(description="Weight in kg at death.", gt=0.0, title="Weight")
    ]


class Height(RootModel[float]):
    root: Annotated[
        float, Field(description="Height in cm at death.", gt=0.0, title="Height")
    ]


class Donor(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    id: Annotated[str, Field(description="Unique donor ID.", title="ID")]
    age: Annotated[
        Age | Literal["<18", ">18"] | None,
        Field(description="Age in years at death.", title="Age"),
    ] = None
    sex: Annotated[Literal["M", "F"] | None, Field(title="Sex")] = None
    weight: Annotated[
        Weight | None, Field(description="Weight in kg at death.", title="Weight")
    ] = None
    height: Annotated[
        Height | None, Field(description="Height in cm at death.", title="Height")
    ] = None
    cause_of_death: Annotated[str | None, Field(title="Cause of death")] = None
    date_of_death: Annotated[date_aliased | None, Field(title="Date of death")] = None
    medical_history: Annotated[str | None, Field(title="Medical history")] = None
    diabetes: Annotated[
        Literal["Type 2", "No", "Yes"] | None, Field(title="Diabetes history")
    ] = None
    hypertension: Annotated[
        Literal["Yes", "No"] | None, Field(title="Hypertension history")
    ] = None
    smoker: Annotated[
        Literal["Never", "Yes"] | None, Field(title="Smoking history")
    ] = None


class Proposal(BaseModel):
    proposal_number: Annotated[
        Literal["md1252", "md1290", "md1389"],
        Field(description="Proposal number.", title="Proposal Number"),
    ]
    title: Annotated[str, Field(description="Proposal title.", title="Title")]
    proposers: Annotated[
        list[str], Field(description="Proposal proposers.", title="Proposers")
    ]


class Registration(BaseModel):
    source_dataset: Annotated[
        str,
        Field(
            description="Name of dataset that is registered (this dataset).",
            title="Source Dataset",
        ),
    ]
    target_dataset: Annotated[
        str,
        Field(
            description="Name of dataset that this dataset is registered to.",
            title="Target Dataset",
        ),
    ]
    translation: Annotated[
        list,
        Field(
            description="Translation vector.",
            max_length=3,
            min_length=3,
            title="Translation",
        ),
    ]
    rotation: Annotated[
        float, Field(description="Rotation in degrees about z-axis", title="Rotation")
    ]
    scale: Annotated[
        float, Field(description="Isotropic scaling factor.", title="Scale")
    ]


class PostMortemIntervalHours(RootModel[float]):
    root: Annotated[
        float,
        Field(
            description="Interval between death and fixation, in hours.",
            ge=0.0,
            title="Post-mortem interval",
        ),
    ]


class Sample(BaseModel):
    organ: Annotated[
        Literal[
            "brain",
            "breast",
            "clitoris",
            "colon",
            "eye",
            "femaleGenitalia",
            "kidney",
            "liver",
            "lung",
            "heart",
            "oesophagus",
            "spinalCord",
            "spine",
            "spleen",
            "heart+lungs",
            "pancreas",
            "placenta",
            "prostate",
            "testis",
            "uterus",
            "thoracicBlock",
            "soleus",
            "achillesTendon",
            "knee",
            "tibia",
        ],
        Field(description="Organ name.", title="Organ"),
    ]
    organ_context: Annotated[
        str | None,
        Field(
            description="Human-readable context for location of the dataset within the whole organ. Only present for some zoom datasets.",
            examples=["upper_lobe", "left", "tumor"],
            title="Organ Context",
        ),
    ] = None
    post_mortem_interval_hours: Annotated[
        PostMortemIntervalHours | None,
        Field(
            description="Interval between death and fixation, in hours.",
            title="Post-mortem interval",
        ),
    ] = None
    fixation_method: Annotated[
        Literal["immersion", "perfusion"] | None,
        Field(description="Fixation preparation method.", title="Fixation Method"),
    ] = None
    fixation_medium: Annotated[
        str | None, Field(description="Fixation medium", title="Fixation Medium")
    ] = None
    organ_infilled: Annotated[bool | str | None, Field(title="Organ Infilled")] = None
    stabilisation_medium: Annotated[
        Literal["agar cubes", "crushed agar", "beads"] | None,
        Field(title="Stabilisation Medium"),
    ] = None
    degassing_method: Annotated[
        Literal["inline", "vacuum", "vacuum+inline"] | None,
        Field(title="Degassing Method"),
    ] = None
    scan_solvent: Annotated[
        Literal["formalin", "ethanol"] | None,
        Field(description="Scan medium solvent.", title="Scan Solvent"),
    ] = None
    scan_solvent_concentration: Annotated[
        float | None,
        Field(
            description="Scan medium solvent concentration in percent.",
            title="Solvent concentration",
        ),
    ] = None
    scan_temperature: Annotated[
        str | float | None,
        Field(
            description="Temperature during scan, in degees celsius (if a number).",
            title="Scan Temperature",
        ),
    ] = "room temperature"


class Energy(RootModel[float]):
    root: Annotated[
        float,
        Field(
            description="Energy used during the scan, measured in keV.",
            gt=0.0,
            title="Energy",
        ),
    ]


class CurrentStart(RootModel[float]):
    root: Annotated[
        float,
        Field(
            description="Current of the synchrotron at the start of the scan, measured in mA.",
            gt=0.0,
            title="Current Start",
        ),
    ]


class NFrames(RootModel[int]):
    root: Annotated[
        int,
        Field(
            description="Number of frames acquired during the scan.",
            gt=0,
            title="Number of frames",
        ),
    ]


class NRef(RootModel[int]):
    root: Annotated[
        int,
        Field(
            description="Number of reference images collected during the scan.",
            ge=0,
            title="Number of references",
        ),
    ]


class NDark(RootModel[int]):
    root: Annotated[
        int,
        Field(
            description="Number of dark images (without illumination) collected during the scan.",
            ge=0,
            title="Number of darks",
        ),
    ]


class LatencyTime(RootModel[float]):
    root: Annotated[
        float,
        Field(
            description="Inactive time between active sensor times, measured in seconds.",
            gt=0.0,
            title="Latency Time",
        ),
    ]


class ExposureTime(RootModel[float]):
    root: Annotated[
        float,
        Field(
            description="Exposure time of a single frame, measured in seconds. This is the sub-frame time multiplied by the number of sub-frames.",
            gt=0.0,
            title="Exposure Time",
        ),
    ]


class SubframeTime(RootModel[float]):
    root: Annotated[
        float,
        Field(
            description="Exposure time of a single sub-frame, measured in seconds.",
            gt=0.0,
            title="Subframe Time",
        ),
    ]


class NSubframes(RootModel[int]):
    root: Annotated[
        int,
        Field(
            description="Number of frames accumulated per exposure.",
            gt=0,
            title="Number of sub-frames",
        ),
    ]


class NScans(RootModel[int]):
    root: Annotated[
        int,
        Field(
            description="Total number of scans. For a helical scan this is always 1. For a zseries this is the total number of scans concatenated to make the full data volume.",
            gt=0,
            title="Number of scans",
        ),
    ]


class ScanTime(RootModel[float]):
    root: Annotated[
        float,
        Field(
            description="Total duration of a single scan in seconds.",
            gt=0.0,
            title="Scan Time",
        ),
    ]


class OpticMagnification(RootModel[float]):
    root: Annotated[
        float,
        Field(
            description="Magnification in the optical setup.",
            gt=0.0,
            title="Optic Magnification",
        ),
    ]


class DistanceSourceSample(RootModel[float]):
    root: Annotated[
        float,
        Field(
            description="Distance between the source and the sample, measured in millimeters.",
            gt=0.0,
            title="Source-sample distance",
        ),
    ]


class DistanceSampleDetector(RootModel[float]):
    root: Annotated[
        float,
        Field(
            description="Distance between the sample and the detector, measured in millimeters.",
            gt=0.0,
            title="Sample-detector distance",
        ),
    ]


class SensorRoiXSize(RootModel[int]):
    root: Annotated[
        int,
        Field(
            description="Number of pixels in the x-dimension of the sensor's region of interest (ROI).",
            gt=0,
            title="Sensor Roi X Size",
        ),
    ]


class SensorRoiYSize(RootModel[int]):
    root: Annotated[
        int,
        Field(
            description="Number of pixels in the y-dimension of the sensor's region of interest (ROI).",
            gt=0,
            title="Sensor Roi Y Size",
        ),
    ]


class SensorBinning(RootModel[list]):
    root: Annotated[
        list,
        Field(
            description="Pixel binning on the sensor.",
            max_length=2,
            min_length=2,
            title="Sensor Binning",
        ),
    ]


class XrayMagnification(RootModel[float]):
    root: Annotated[
        float,
        Field(
            description="Magnification of X-rays due to divergence between the source and detector.",
            gt=0.0,
            title="Xray Magnification",
        ),
    ]


class Scan(BaseModel):
    date: Annotated[
        date_aliased,
        Field(description="Date when the scan was performed.", title="Date"),
    ]
    beamline: Annotated[
        Literal["BM05", "BM18"],
        Field(description="ESRF beamline where scan was performed.", title="Beamline"),
    ]
    energy: Annotated[
        Energy | None,
        Field(
            description="Energy used during the scan, measured in keV.", title="Energy"
        ),
    ] = None
    current_start: Annotated[
        CurrentStart | None,
        Field(
            description="Current of the synchrotron at the start of the scan, measured in mA.",
            title="Current Start",
        ),
    ] = None
    filling_mode: Annotated[
        Literal[
            "16 bunch",
            "7/8 multibunch",
            "uniform multibunch",
            "28*12+1 bunch",
            "4 bunch",
            "2*1/3 multibunch",
        ]
        | None,
        Field(
            description="Mode of filling used in the synchrotron storage ring.",
            title="Filling Mode",
        ),
    ] = None
    n_frames: Annotated[
        NFrames | None,
        Field(
            description="Number of frames acquired during the scan.",
            title="Number of frames",
        ),
    ] = None
    n_ref: Annotated[
        NRef | None,
        Field(
            description="Number of reference images collected during the scan.",
            title="Number of references",
        ),
    ] = None
    n_dark: Annotated[
        NDark | None,
        Field(
            description="Number of dark images (without illumination) collected during the scan.",
            title="Number of darks",
        ),
    ] = None
    latency_time: Annotated[
        LatencyTime | None,
        Field(
            description="Inactive time between active sensor times, measured in seconds.",
            title="Latency Time",
        ),
    ] = None
    exposure_time: Annotated[
        ExposureTime | None,
        Field(
            description="Exposure time of a single frame, measured in seconds. This is the sub-frame time multiplied by the number of sub-frames.",
            title="Exposure Time",
        ),
    ] = None
    subframe_time: Annotated[
        SubframeTime | None,
        Field(
            description="Exposure time of a single sub-frame, measured in seconds.",
            title="Subframe Time",
        ),
    ] = None
    n_subframes: Annotated[
        NSubframes | None,
        Field(
            description="Number of frames accumulated per exposure.",
            title="Number of sub-frames",
        ),
    ] = None
    scan_type: Annotated[
        Literal["helical", "zseries"],
        Field(description="Type of scan.", title="Scan Type"),
    ]
    scan_range: Annotated[
        float,
        Field(
            description="Angular range of a single scan in degrees.",
            gt=0.0,
            title="Scan Range",
        ),
    ]
    n_scans: Annotated[
        NScans | None,
        Field(
            description="Total number of scans. For a helical scan this is always 1. For a zseries this is the total number of scans concatenated to make the full data volume.",
            title="Number of scans",
        ),
    ] = None
    acquisition: Annotated[
        Literal["half", "quarter"] | None,
        Field(description="Type of tomographic acquisition.", title="Acquisition"),
    ] = None
    z_step: Annotated[
        float | None,
        Field(description="Displacement in millimeters between scans.", title="Z Step"),
    ] = None
    scan_time: Annotated[
        ScanTime | None,
        Field(
            description="Total duration of a single scan in seconds.", title="Scan Time"
        ),
    ] = None
    filters: Annotated[
        list[str] | None,
        Field(description="List of filters used during the scan.", title="Filters"),
    ] = None
    scintillator: Annotated[
        Literal[
            "LuAG:Ce 1000um",
            "LuAG:Ce 100um",
            "LuAG:Ce 100um, reflective",
            "LuAG:Ce 100um, lead glass meniscus",
            "LuAG:Ce 2000um",
            "LuAG:Ce 2000um, reflective",
            "Multi LuAG:Ce 2000um, reflective",
            "LuAG:Ce 200um",
            "LuAG:Ce 250um",
            "LuAG:Ce 250um, reflective",
            "LuAG:Ce 25um",
            "LuAG:Ce 50um",
            "LuAG:Ce 50um, reflective",
            "LuAG:Ce 500um",
            "LySO:Ce 8.5um",
            "LySO:Ce 23um",
            "LySO:Ce 2000um, reflective",
            "Gadox 60um",
            "CSI trixell structured 650um, 5um pitch",
            "GGG 8um",
            "GGG 20um",
            "GAGG 1000um, reflective",
            "GAGG 100um, reflective",
            "GAGG 50um, reflective",
            "Dual LySO:Ce 5000um, reflective",
        ]
        | None,
        Field(
            description="Scintillator used for converting X-rays to visible light.",
            title="Scintillator",
        ),
    ] = None
    optic: Annotated[
        Literal[
            "Zoom",
            "DZoom",
            "10x 0.4NA reflective optic",
            "Hasselblad tandem optic 100mm/300mm",
            "Hasselblad revolved 100 100",
            "LAFIP2 optic with canon 50mm",
            "Fixed x10",
            "Fixed x2.85",
            "Fixed x2",
            "Fixed x1",
            "Fixed x0.5",
            "Fixed x0.125",
            "Fixed x0.1",
            "Twinmic 5",
            "Twinmic 10",
            "Twinmic 20",
            "Triplemic 10",
        ]
        | None,
        Field(description="Optical components used during the scan", title="Optic"),
    ] = None
    optic_magnification: Annotated[
        OpticMagnification | None,
        Field(
            description="Magnification in the optical setup.",
            title="Optic Magnification",
        ),
    ] = None
    distance_source_sample: Annotated[
        DistanceSourceSample | None,
        Field(
            description="Distance between the source and the sample, measured in millimeters.",
            title="Source-sample distance",
        ),
    ] = None
    distance_sample_detector: Annotated[
        DistanceSampleDetector | None,
        Field(
            description="Distance between the sample and the detector, measured in millimeters.",
            title="Sample-detector distance",
        ),
    ] = None
    sensor_name: Annotated[
        Literal["PCO edge 4.2 CLHS", "Iris15"] | None,
        Field(
            description="Name of the sensor used during the scan.", title="Sensor Name"
        ),
    ] = None
    sensor_mode: Annotated[
        Literal["rolling shutter", "FFM"] | None,
        Field(
            description="Operating mode of the sensor during the scan.",
            title="Sensor Mode",
        ),
    ] = None
    sensor_roi_x_size: Annotated[
        SensorRoiXSize | None,
        Field(
            description="Number of pixels in the x-dimension of the sensor's region of interest (ROI).",
            title="Sensor Roi X Size",
        ),
    ] = None
    sensor_roi_y_size: Annotated[
        SensorRoiYSize | None,
        Field(
            description="Number of pixels in the y-dimension of the sensor's region of interest (ROI).",
            title="Sensor Roi Y Size",
        ),
    ] = None
    sensor_binning: Annotated[
        SensorBinning | None,
        Field(description="Pixel binning on the sensor.", title="Sensor Binning"),
    ] = None
    pixel_size: Annotated[
        float,
        Field(
            description="Size of a pixel at the sample, measured in micrometers.",
            gt=0.0,
            title="Pixel Size",
        ),
    ]
    xray_magnification: Annotated[
        XrayMagnification | None,
        Field(
            description="Magnification of X-rays due to divergence between the source and detector.",
            title="Xray Magnification",
        ),
    ] = None
    technique: Annotated[
        Literal["Hierarchical Phase-Contrast Tomography (HiP-CT)"],
        Field(title="Experimental technique."),
    ] = "Hierarchical Phase-Contrast Tomography (HiP-CT)"
    experiment_type: Annotated[str | None, Field(title="Experiment Type")] = (
        "Tomography"
    )


class Citation(BaseModel):
    title: Annotated[str, Field(description="Citation title", title="Title")]
    contributors: Annotated[
        list[Contributor],
        Field(
            description="Dataset contributors, listed alphabetically by name.",
            title="Contributors",
        ),
    ]
    doi: Annotated[str | None, Field(description="Dataset DOI", title="Doi")] = None
    author_list: Annotated[
        list[str],
        Field(
            description="Author list. If citing this data, please use this authorship list in the order given",
            title="Author List",
        ),
    ]


class HOAMetadata(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    name: Annotated[str, Field(description="Unique name for dataset.", title="Name")]
    dataset_type: Annotated[
        Literal["overview", "zoom"],
        Field(
            description="Type of dataset. Overview datasets contain the full sample. Zoom datasets contain a zoom into a particular volume of interest within the sample.",
            title="Dataset Type",
        ),
    ]
    voi: Annotated[
        str,
        Field(
            description="Volume of interest name. For zoom datasets this gives a unique identifier among zoom datasets taken within the same sample.",
            title="Volume of Interest",
        ),
    ]
    data: Annotated[Data, Field(description="Data metadata")]
    sample: Annotated[Sample, Field(description="Sample metadata")]
    donor: Annotated[Donor, Field(description="Donor metadata")]
    scan: Annotated[Scan, Field(description="Scan metadata")]
    proposal: Annotated[Proposal, Field(description="Beamtime proposal metadata")]
    registration: Annotated[
        Registration | None,
        Field(
            description="Metadata mapping high resolution volume of interest datasets to a full organ dataset. Not present for full organ datasets.",
            title="Registration parameters",
        ),
    ] = None
    citation: Annotated[Citation, Field(title="Dataset citation")]
    metadata_version: Annotated[
        Literal["1.1"],
        Field(description="Version of this metadata file.", title="Metadata Version"),
    ] = "1.1"
